<?php
/**
 * Created by PhpStorm.
 * User: crose
 * Date: 1/4/16
 * Time: 7:14 PM
 */

namespace qfq;

use qfq;
use qfq\DbException;
use qfq\Store;
use qfq\CodeException;
use qfq\UserException;
use qfq\Support;


require_once(__DIR__ . '/exceptions/UserException.php');
require_once(__DIR__ . '/exceptions/CodeException.php');
require_once(__DIR__ . '/exceptions/DbException.php');

require_once(__DIR__ . '/store/Store.php');
require_once(__DIR__ . '/helper/Support.php');
require_once(__DIR__ . '/helper/BindParam.php');

/**
 * Class Database
 * @package qfq
 */
class Database {

    /**
     * @var Store
     */
    private $store = null;

    /**
     * @var \mysqli
     */
    private $mysqli = null;

    /**
     * @var \mysqli_stmt
     */
    private $mysqli_stmt = null;

    /**
     * @var \mysqli_result
     */
    private $mysqli_result = null;


    /**
     * Returns current data base handle from Store[System][SYSTEM_DBH].
     * If not exists: open database and store the new dbh in Store[System][SYSTEM_DBH]
     *
     * @throws CodeException
     * @throws UserException
     */
    public function __construct() {
        $this->store = Store::getInstance();

        if ($this->mysqli === null) {
            $this->mysqli = $this->dbConnect();
        }
    }


    /**
     * Open mysqli database connection if not already done.
     *
     * @return \mysqli
     * @throws UserException
     */
    private function dbConnect() {
        $mysqli = null;

        $dbuser = $this->store->getVar(SYSTEM_DBUSER, STORE_SYSTEM);
        $dbserver = $this->store->getVar(SYSTEM_DBSERVER, STORE_SYSTEM);
        $dbpw = $this->store->getVar(SYSTEM_DBPW, STORE_SYSTEM);
        $db = $this->store->getVar(SYSTEM_DB, STORE_SYSTEM);

        $mysqli = new \mysqli($dbserver, $dbuser, $dbpw, $db);

        if ($mysqli->connect_error) {
            throw new UserException ("Error open Database 'mysql:host=" . $dbserver . ";dbname=" . $db . ";dbuser=" . $dbuser . "'': " . $mysqli->connect_errno . PHP_EOL . $mysqli->connect_error, ERROR_DB_OPEN);
        }

        return $mysqli;
    }

    /**
     * Return the number of rows returned by the last call to execute().
     *
     * If execute() has never been called, returns FALSE.
     *
     * @return mixed Number of rows returned by last call to execute(). If Database::execute()
     *     has never been called prior a call to this method, false is returned.
     */
    public function getRowCount() {
        if ($this->mysqli_result == null) {
            return false;
        }

        return $this->mysqli_result->num_rows;
    }

    /**
     * Get the values for a given ENUM or SET column
     *
     * @param string $table name of the table
     * @param string $columnName name of the column
     *
     * @throws UserException if the table or column does not exist, or is not of type ENUM or SET
     * @return array
     */
    public function getEnumSetValueList($table, $columnName) {

        $columnDefinition = $this->getFieldDefinitionFromTable($table, $columnName);
        $setEnumDefinition = $columnDefinition["Type"];

        // $setEnumDefinition holds now a string like
        // String:  enum('','red','blue','green')
        $len = mb_strlen($setEnumDefinition);

        # "enum('" = 6, "set('" = 5
        $tokenLength = strpos($setEnumDefinition, "'") + 1;

        // count("enum('") == 6, count("')") == 2
        $enumSetString = mb_substr($setEnumDefinition, $tokenLength, $len - (2 + $tokenLength));

        // String: ','red','blue','green

        if (($setEnumValueList = explode("','", $enumSetString)) === false) {
            return array();
        }

        return $setEnumValueList;
    }

    /**
     * Get database column definition.
     *
     * If the column is not found in the table, an exception is thrown.
     *
     * @param string $table name of the table
     *
     * @param string $columnName name of the column
     * @return array the definition of the column as retrieved by Database::getTableDefinition().
     *
     * @throws \qfq\DbException
     */
    private function getFieldDefinitionFromTable($table, $columnName) {
        $tableDefinition = $this->getTableDefinition($table);
        foreach ($tableDefinition AS $row) {
            if ($row["Field"] == $columnName) {
                return $row;
            }
        }
        throw new DbException("Column name '$columnName' not found in table '$table'.", ERROR_DB_COLUMN_NOT_FOUND_IN_TABLE);
    }

    /**
     * Get all column definitions for a table. Return Assoc Array:
     *
     * Field      Type                      Null    Key    Default    Extra
     * --------------------------------------------------------------------------
     * id         bigint(20)                 NO     PRI    NULL    auto_increment
     * name       varchar(128)               YES           NULL
     * firstname  varchar(128)               YES           NULL
     * gender     enum('','male','female')   NO            male
     * groups     set('','a','b','c')        NO            a
     *
     * @param string $table table to retrieve column definition from
     *
     * @return array column definition of table as returned by SHOW FIELDS FROM as associative array.
     */
    public function getTableDefinition($table) {
        return $this->sql("SHOW FIELDS FROM `$table`");
    }

    /**
     * Fires query $sql and fetches result als assoc array (all modes but ROW_KEYS) or as num array (mode: ROW_EKYS). Throws exception.
 * $mode
     *  ROW_REGULAR: Return 2-dimensional assoc array. Every query row is one array row.
     *  ROW_IMPLODE_ALL: Return string. All cells of all rows imploded to one string.
     *  ROW_EXPECT_0: Return empty string if there is now record row, Else an exception.
     *  ROW_EXPECT_1: Return 1-dimensional assoc array if there are exact one row. Else an exception.
     *  ROW_EXPECT_0_1: Return empty string if there is no row. Return 1- dimensional assoc array if there is one row. Else an exception.
     *  ROW_EXPECT_GE_1: Like 'ROW_REGULAR'. Throws an exception if there is an empty resultset.
     *  ROW_KEYS: Return 2-dimensional num(!) array. Every query row is one array row. In $keys are the column names.
     *
     * @param $sql
     * @param string $mode
     * @param array $parameterArray
     * @param string $specificMessage
     * @return mixed|null
     * @throws \qfq\CodeException
     * @throws \qfq\DbException
     */
    public function sql($sql, $mode = ROW_REGULAR, array $parameterArray = array(), $specificMessage = '', array &$keys = array()) {
        $result = array();
        $this->closeMysqliStmt();

        // CR often forgets to specify the $mode and use prepared statement with parameter instead.
        if (is_array($mode))
            throw new CodeException("Probably a parameter forgotten: \$mode ?");

        // for error reporting in exception
        if ($specificMessage)
            $specificMessage .= " ";

        $count = $this->prepareExecute($sql, $parameterArray);
        if ($count === false) {
            throw new DbException($specificMessage . "No idea why this error happens - please take some time and check this: $sql", ERROR_DB_GENERIC_CHECK);
        }

        switch ($mode) {
            case ROW_IMPLODE_ALL:
                $result = $this->fetchAll($mode);
                break;
            case ROW_KEYS:
            case ROW_REGULAR:
                $result = $this->fetchAll($mode, $keys);
                break;
            case ROW_EXPECT_0:
                if ($count === 0)
                    $result = array();
                else
                    throw new DbException($specificMessage . "Expected no row, got $count rows: $sql", ERROR_DB_TOO_MANY_ROWS);
                break;
            case ROW_EXPECT_1:
                if ($count === 1)
                    $result = $this->fetchAll($mode)[0];
                else
                    throw new DbException($specificMessage . "Expected one row, got 0 or more than 1: $sql", ERROR_DB_COUNT_DO_NOT_MATCH);
                break;
            case ROW_EXPECT_0_1:
                if ($count === 1)
                    $result = $this->fetchAll($mode)[0];
                elseif ($count === 0)
                    $result = array();
                else
                    throw new DbException($specificMessage . "Expected no row, got $count rows: $sql", ERROR_DB_TOO_MANY_ROWS);
                break;
            case ROW_EXPECT_GE_1:
                if ($count > 0)
                    $result = $this->fetchAll($mode);
                else
                    throw new DbException($specificMessage . "Expected at least one row, got nothing: $sql", ERROR_DB_TOO_FEW_ROWS);
                break;

            default:
                throw new DbException($specificMessage . "Unknown mode: $mode", ERROR_UNKNOWN_MODE);
        }

        $this->closeMysqliStmt();

        return $result;
    }

    /**
     * Close an optional open MySQLi Statement.
     *
     * @throws \qfq\DbException
     */
    private function closeMysqliStmt() {

        if ($this->mysqli_result !== null && $this->mysqli_result !== false) {
            $this->mysqli_result->free_result();
        }

        if ($this->mysqli_stmt !== null && $this->mysqli_stmt !== false) {
            $this->mysqli_stmt->free_result();
            if (!$this->mysqli_stmt->close())
                throw new DbException('Error closing mysqli_stmt' . ERROR_DB_CLOSE_MYSQLI_STMT);
        }
        $this->mysqli_stmt = null;
        $this->mysqli_result = null;
    }

    /**
     * Execute a prepared SQL statement like SELECT, INSERT, UPDATE, DELETE, SHOW, ...
     *
     * Returns the number of selected rows (SELECT, SHOW, ..) or the affected rows (UPDATE) or the last insert id (INSERT)
     *
     * @param string $sql SQL statement with prepared statement variable.
     * @param array $parameterArray parameter array for prepared statement execution.
     * @return int|mixed
     * @throws \qfq\CodeException
     * @throws \qfq\DbException
     * @throws \qfq\UserException
     */
    private function prepareExecute($sql, array $parameterArray = array()) {
        $result = 0;
        $this->store->setVar(SYSTEM_SQL_FINAL, $sql, STORE_SYSTEM);
        $this->store->setVar(SYSTEM_SQL_PARAM_ARRAY, $parameterArray, STORE_SYSTEM);

        // Logfile
        $this->dbLog($sql, $parameterArray);

        if (false === ($this->mysqli_stmt = $this->mysqli->prepare($sql))) {
            throw new DbException('[ mysqli: ' . $this->mysqli->errno . ' ] ' . $this->mysqli->error, ERROR_DB_PREPARE);
        }


        if (count($parameterArray) > 0) {
            if (false === $this->prepareBindParam($parameterArray)) {
                throw new DbException('[ mysqli: ' . $this->mysqli_stmt->errno . ' ] ' . $this->mysqli_stmt->error, ERROR_DB_BIND);
            }
        }

        if (false === $this->mysqli_stmt->execute()) {
            throw new DbException('[ mysqli: ' . $this->mysqli_stmt->errno . ' ] ' . $this->mysqli_stmt->error, ERROR_DB_EXECUTE);
        }

        $msg = '';
        $count = 0;
        $command = strtoupper(explode(' ', $sql, 2)[0]);
        switch ($command) {
            case 'SELECT':
            case 'SHOW':
            case 'DESCRIBE':
            case 'EXPLAIN':
                if (false === ($result = $this->mysqli_stmt->get_result())) {
                    throw new DbException('[ mysqli: ' . $this->mysqli_stmt->errno . ' ] ' . $this->mysqli_stmt->error, ERROR_DB_EXECUTE);
                }
                $this->mysqli_result = $result;
                $count = $this->mysqli_result->num_rows;
                $msg = 'Get rows: ' . $count;
                break;
            case 'INSERT':
                $count = $this->mysqli->insert_id;
                $msg = 'ID: ' . $count;
                break;
            case 'UPDATE':
            case 'REPLACE':
            case 'DELETE':
                $count = $this->mysqli->affected_rows;
                $msg = 'Affected rows: ' . $count;
                break;
            default:
                break;
        }

        $this->store->setVar(SYSTEM_SQL_COUNT, $count, STORE_SYSTEM);

        // Logfile
        $this->dbLog($msg);

        return $count;
    }

    /**
     * @param $sql
     * @param $parameterArray
     * @throws \qfq\CodeException
     */
    private function dbLog($sql, $parameterArray = array()) {
        $msg = '[' . date('r') . '][';

        if (count($parameterArray) === 0) {
            $msg .= $sql;
        } else {

            $sqlArray = explode('?', $sql);
            $ii = 0;
            foreach ($parameterArray as $value) {
                if (isset($sqlArray[$ii])) {
                    $msg .= $sqlArray[$ii++] . "'" . $value . "'";
                } else {
                    $msg = '?';
                }
            }
            if (isset($sqlArray[$ii]))
                $msg .= $sqlArray[$ii];
        }

        $msg .= ']';

        Support::log($msg);
    }

    /**
     * @param $arr
     */
    private function prepareBindParam($arr) {

        $bindParam = new BindParam();

        for ($ii = 0; $ii < count($arr); $ii++) {
            $bindParam->add($arr[$ii]);
        }
        call_user_func_array([$this->mysqli_stmt, 'bind_param'], $bindParam->get());
    }

    /**
     * Fetch all rows of the result as associative array.
     *
     * mode:
     *  ROW_IMPLODE_ALL: Return string. All cells of all rows imploded to one string.
     *  ROW_KEYS: Retrun num array with column names in $keys
     *  default: Return 2-dimensional assoc array
     *
     * @param string $mode
     * @param array $keys
     * @return array|bool|mixed|string false in case of error.
     *              Empty string is returned if the query didn't yield any rows.
     *              All rows as Multi Assoc array if $mode!=IMPLODE_ALL.
     *              All rows and all columns imploded to one string if $mode=IMPLODE_ALL
     *
     */
    private function fetchAll($mode = '', &$keys = array()) {
        if ($this->mysqli_result == null || $this->mysqli_result == false) {
            return false;
        }

        if ($this->mysqli_result->num_rows === 0) {
            return ($mode === ROW_IMPLODE_ALL) ? "" : array();
        }

        switch ($mode) {
            case ROW_IMPLODE_ALL:
                $str = "";
                foreach ($this->mysqli_result->fetch_all(MYSQLI_ASSOC) as $row) {
                    $str .= implode($row);
                }
                return $str;
                break;

            case ROW_KEYS:
                $keys = array();

                for ($ii = 0; $ii < $this->mysqli_result->field_count; $ii++) {
                    $keys[$ii] = $this->mysqli_result->fetch_field_direct($ii)->name;
                }
                return $this->mysqli_result->fetch_all(MYSQLI_NUM);
                break;

            default:
                return $this->mysqli_result->fetch_all(MYSQLI_ASSOC);
        }
    }

    /**
     * Wrapper for sql(), to simplyfy access.
     *
     * @param $sql
     * @param array $keys
     * @return array|bool
     * @throws \qfq\DbException
     */
    public function sqlKeys($sql, array &$keys) {

        return $this->sql($sql, ROW_KEYS, array(), '', $keys);
    }

    /**
     * Returns lastInsertId
     *
     * @return string
     */
    public function getLastInsertId() {
        // Do not try to use $this->mysqli->lastInsertId - this is not valid at any given time.
        return $this->mysqli->insert_id;
    }

}