<?php
/**
 * @author Rafael Ostertag <rafael.ostertag@math.uzh.ch>
 */

namespace qfq;

use qfq\KeyValueStringParser;

require_once(__DIR__ . '/../../qfq/Constants.php');


require_once(__DIR__ . '/../../qfq/helper/KeyValueStringParser.php');

class KeyValueStringParserTest extends \PHPUnit_Framework_TestCase {

    public function testSingleKeyValuePair() {
        $actual = keyValueStringParser::parse("key:value");

        $this->assertCount(1, $actual);
        $this->assertArrayHasKey('key', $actual);
        $this->assertEquals('value', $actual['key']);
    }

    public function testKeyWithoutValue() {
        $actual = keyValueStringParser::parse("keywithoutvalue");
        $this->assertCount(1, $actual);
        $this->assertArrayHasKey('keywithoutvalue', $actual);
        $this->assertSame("", $actual['keywithoutvalue']);
    }

    public function testGOODNAMEHERE() {
        $actual = keyValueStringParser::parse(",,");

        $this->assertCount(0, $actual);
    }

    /**
     * @expectedException \qfq\UserException
     */
    public function testNoKey() {
        keyValueStringParser::parse(":value,key:value");
    }

    public function testNoValue() {
        $actual = keyValueStringParser::parse("key1:,key2:value2");
        $this->assertCount(2, $actual);
        $this->assertArrayHasKey('key1', $actual);
        $this->assertArrayHasKey('key2', $actual);
        $this->assertEquals('', $actual['key1']);
        $this->assertEquals('value2', $actual['key2']);
    }

    public function testEmptyKeyValuePairString() {
        $actual = keyValueStringParser::parse("");

        $this->assertCount(0, $actual);
    }

    public function testMultipleKeyValuePairs() {
        $actual = keyValueStringParser::parse("key1:value1,key2:value2");

        $this->assertCount(2, $actual);
        $this->assertArrayHasKey('key1', $actual);
        $this->assertArrayHasKey('key2', $actual);
        $this->assertEquals('value1', $actual['key1']);
        $this->assertEquals('value2', $actual['key2']);
    }

    public function testKeyValueSeparatorInValue() {
        $actual = keyValueStringParser::parse("key1:val:ue1,key2:value2");

        $this->assertCount(2, $actual);
        $this->assertArrayHasKey('key1', $actual);
        $this->assertArrayHasKey('key2', $actual);
        $this->assertEquals('val:ue1', $actual['key1']);
        $this->assertEquals('value2', $actual['key2']);
    }

    public function testWhiteSpaceHandling() {
        $actual = keyValueStringParser::parse(" key1 : val:ue1 , key2 : value2 ");

        $this->assertCount(2, $actual);
        $this->assertArrayHasKey('key1', $actual);
        $this->assertArrayHasKey('key2', $actual);
        $this->assertEquals('val:ue1', $actual['key1']);
        $this->assertEquals('value2', $actual['key2']);
    }

    public function testSourroundingQuotes() {
        $actual = keyValueStringParser::parse("key1:\" val:ue1 \", key2:' value2 ', key3:\"value3', key4:''");

        $expected = [
            'key1' => ' val:ue1 ',
            'key2' => ' value2 ',
            'key3' => "\"value3'",
            'key4' => ''
        ];
        $this->assertEquals($expected, $actual);
    }

    public function testComments() {
        $actual = keyValueStringParser::parse(" key1 : val:ue1 , # key2 : value2 , ; : broken key value in comment, key3 : valid ");

        $expected = [
            'key1' => 'val:ue1',
            'key3' => "valid",
        ];
        $this->assertEquals($expected, $actual);
    }

    public function testUnparse() {
        $array = keyValueStringParser::parse("key1:\" val:ue1 \", key2:' value2 ', key3:\"value3'");
        $actual = keyValueStringParser::unparse($array);
        $expected = "key1:\" val:ue1 \",key2:\" value2 \",key3:\"value3'";

        $this->assertSame($expected, $actual);

    }

    public function testKeyValuePairCR() {

        $actual = keyValueStringParser::parse("key1=value1\nkey2=value2", "=", "\n");

        $this->assertCount(2, $actual);
        $this->assertArrayHasKey('key1', $actual);
        $this->assertArrayHasKey('key2', $actual);
        $this->assertEquals('value1', $actual['key1']);
        $this->assertEquals('value2', $actual['key2']);
    }

    public function testParseKeyValueSingle() {
        $actual = keyValueStringParser::parse('value1,value2', ':', ',', KVP_IF_VALUE_EMPTY_COPY_KEY);
        $expected = [
            'value1' => 'value1',
            'value2' => 'value2',
        ];
        $this->assertEquals($expected, $actual);
        $this->assertCount(2, $actual);

        $actual = keyValueStringParser::parse('key1:value1,key2:value2', ':', ',', KVP_IF_VALUE_EMPTY_COPY_KEY);
        $expected = [
            'key1' => 'value1',
            'key2' => 'value2',
        ];
        $this->assertEquals($expected, $actual);
        $this->assertCount(2, $actual);

        $actual = keyValueStringParser::parse('key1:value1,key2:value2', ':', ',', KVP_VALUE_GIVEN);
        $expected = [
            'key1' => 'value1',
            'key2' => 'value2',
        ];
        $this->assertEquals($expected, $actual);
        $this->assertCount(2, $actual);

        $actual = keyValueStringParser::parse('value1,value2', ':', ',', KVP_VALUE_GIVEN);
        $expected = [
            'value1' => '',
            'value2' => '',
        ];
        $this->assertEquals($expected, $actual);
        $this->assertCount(2, $actual);
    }

    public function testParseDefaults() {
        $actual = keyValueStringParser::parse("key1=value1,key2=value2");
        $expected = keyValueStringParser::parse("key1=value1,key2=value2", ":", ",");
        $this->assertEquals($expected, $actual);
        $this->assertCount(2, $actual);
    }
}
